#!/uar/bin/perl -w
#
# This is quite basic, but I can't make up anything to test at the 
# moment 


use strict;
use lib 't/lib';
use Test::More tests => 1;

use_ok('CPANPLUS::Shell::Curses');

exit;
